#include <avr/io.h>
#include "tvtext.h"

; Registers.
#define r_font_row         r15 /* Stores a row of pixels from the font data temporarily. */
#define r_characters_left  r16 /* Stores the number of characters left to draw on a scanline. */
#define r_font_data_l      r17 /* Points to the LSB of the font data table, offset by the current scanline. */
#define r_font_data_h      r18 /* Points to the MSB of the font data table, offset by the current scanline. */

#define t_short_vsync        4 /* Length of the "short" sync pulses. */
#define t_long_vsync        28 /* Length of the "long" sync pulses. */
#define t_active_sync        4 /* Length of the sync pulse on active display. */
#define t_active_back_porch  4 /* Length of the back porch on active display. */
#define t_active_picture    56 /* Length of the picture data on active display. */

; Variables.
.lcomm timer1_compa_handler, 2   ; Function pointer to current TIMER1_COMPA_vect handler.
.lcomm scan_region_repeat, 2     ; Used to count down display regions when rasterising.
.lcomm cursor_cell, 2            ; Pointer to the buffer cell that contains the cursor.
.lcomm cursor_char, 1            ; Character to draw in the cursor's cell.

#define cycles_us(us) (us)*20-1

.global tvtext_driver_init
tvtext_driver_init:
	push r20
	ldi r20, lo8(pm(long_vsync_sync))
	sts timer1_compa_handler+0, r20
	ldi r20, hi8(pm(long_vsync_sync))
	sts timer1_compa_handler+1, r20
	ldi r20, 5
	sts scan_region_repeat, r20
	pop r20
	ret

.global TIMER1_COMPA_vect
TIMER1_COMPA_vect:
	in r_sreg_save,_SFR_IO_ADDR(SREG) ; [1]
	push ZL                           ; [2]
	push ZH                           ; [2]
	push YL                           ; [2]
	push YH                           ; [2]

	; Disable CTC.
	ldi ZL, _BV(CS10)
	sts _SFR_MEM_ADDR(TCCR1B),ZL

	lds ZL,_SFR_MEM_ADDR(TCNT1L) ; [2]
	sbrs ZL,0                    ; [2/1] \_ bit 0 = 0: 3cc \  00 -> 7
	rjmp .                       ; [2]   /  bit 0 = 1: 2cc  | 01 -> 6
	sbrs ZL,1                    ; [2/1] \_ bit 1 = 0: 4cc  | 10 -> 5
	lpm ZL,Z                     ; [3]   /  bit 1 = 1: 2cc /  11 -> 4

	lds ZL, timer1_compa_handler+0      ; [2]
	lds ZH, timer1_compa_handler+1      ; [2]
	ijmp                                ; [2]

exit_timer1_compa_vect:
	sts timer1_compa_handler+0, ZL
	sts timer1_compa_handler+1, ZH
	sts _SFR_MEM_ADDR(OCR1AH), YH
	sts _SFR_MEM_ADDR(OCR1AL), YL

	; Re-enable CTC.
	ldi ZL, _BV(WGM12) | _BV(CS10)
	sts _SFR_MEM_ADDR(TCCR1B),ZL

	; Clear interrupts (if any happened).
	sbi _SFR_IO_ADDR(TIFR1), OCF1A

	pop YH
	pop YL
	pop ZH
	pop ZL
	out _SFR_IO_ADDR(SREG),r_sreg_save ; [1]
	reti

; 30uS of sync.
long_vsync_sync:
	cbi _SFR_IO_ADDR(TVTEXT_SYNC_PORT), TVTEXT_SYNC_BIT
	ldi YL,lo8(cycles_us(t_long_vsync))
	ldi YH,hi8(cycles_us(t_long_vsync))
	ldi ZL,lo8(pm(long_vsync_black))
	ldi ZH,hi8(pm(long_vsync_black))
	jmp exit_timer1_compa_vect

; 2uS of black.
long_vsync_black:
	sbi _SFR_IO_ADDR(TVTEXT_SYNC_PORT), TVTEXT_SYNC_BIT

	; Is this the last long vsync?
	lds YL, scan_region_repeat
	dec YL
	brne long_vsync_not_finished
	; Long sync is always followed by 5x short sync.
	ser YL
	sts scan_region_repeat+1, YL ; Used to denote that we're entering short sync from long sync.
	ldi YL, 5
	ldi ZL,lo8(pm(short_vsync_sync))
	ldi ZH,hi8(pm(short_vsync_sync))
	jmp long_vsync_finished
long_vsync_not_finished:
	ldi ZL,lo8(pm(long_vsync_sync))
	ldi ZH,hi8(pm(long_vsync_sync))
long_vsync_finished:
	sts scan_region_repeat, YL
	ldi YL,lo8(cycles_us(t_short_vsync))
	ldi YH,hi8(cycles_us(t_short_vsync))
	jmp exit_timer1_compa_vect

; 2uS of sync.
short_vsync_sync:
	cbi _SFR_IO_ADDR(TVTEXT_SYNC_PORT),TVTEXT_SYNC_BIT
	ldi YL,lo8(cycles_us(t_short_vsync))
	ldi YH,hi8(cycles_us(t_short_vsync))
	ldi ZL,lo8(pm(short_vsync_black))
	ldi ZH,hi8(pm(short_vsync_black))
	jmp exit_timer1_compa_vect

; 30uS of black.
short_vsync_black:
	sbi _SFR_IO_ADDR(TVTEXT_SYNC_PORT), TVTEXT_SYNC_BIT
	; Is this the last short vsync?
	lds YL,scan_region_repeat
	dec YL
	brne short_vsync_not_finished
	; Short sync is either followed by 5x long sync or active display.
	lds YL,scan_region_repeat+1
	inc YL
	breq short_vsync_enter_active_display
short_vsync_enter_long_sync:
	; We've just finished a frame!
	lds ZL,tvtext_frame_counter+0
	inc ZL
	sts tvtext_frame_counter+0,ZL
	brne 1f
	lds ZH,tvtext_frame_counter+1
	inc ZH
	sts tvtext_frame_counter+1,ZH
1:	ldi YL, 5
	ldi ZL,lo8(pm(long_vsync_sync))
	ldi ZH,hi8(pm(long_vsync_sync))
	jmp short_vsync_finished
short_vsync_enter_active_display:
	; We're about to enter active display.
	call set_up_cursor
	; Initialise the scanline counter, skipping the first few lines.
	ldi YL,lo8(-(32+1))
	sts tvtext_scanline_counter+0, YL
	ldi YH,hi8(-(32+1))
	sts tvtext_scanline_counter+1, YH
	; 304 lines in PAL.
	ldi YH,hi8(-304)
	sts scan_region_repeat+1, YH
	ldi YL,lo8(-304)
	ldi ZL,lo8(pm(active_sync))
	ldi ZH,hi8(pm(active_sync))
	jmp short_vsync_finished
short_vsync_not_finished:
	ldi ZL,lo8(pm(short_vsync_sync))
	ldi ZH,hi8(pm(short_vsync_sync))
short_vsync_finished:
	sts scan_region_repeat, YL
	ldi YL,lo8(cycles_us(t_long_vsync))
	ldi YH,hi8(cycles_us(t_long_vsync))
	jmp exit_timer1_compa_vect

set_up_cursor:
	; Calculate the cursor offset and character.
	push r0
	push r1
	lds r0,tvtext_cursor_row
	ldi YL,32
	mul r0,YL
	lds YL,tvtext_cursor_column
	clr YH
	add r0,YL
	adc r1,YH
	ldi YL,lo8(tvtext_buffer)
	ldi YH,hi8(tvtext_buffer)
	add r0,YL
	adc r1,YH
	sts cursor_cell+0,r0
	sts cursor_cell+1,r1
	movw YL,r0
	ld YL,Y
	lds YH,tvtext_flags
	; Is the cursor visible?
	sbrs YH,TVTEXT_CURSOR_ENABLED
	rjmp found_cursor_char
	sbrs YH,TVTEXT_CURSOR_VISIBLE
	rjmp found_cursor_char
	lds YL,tvtext_cursor
found_cursor_char:
	sts cursor_char,YL
	; Do we need to blink the cursor?
	lds YL,tvtext_cursor_flash_timer
	dec YL
	brne no_flash_cursor
	; Blink the cursor.
	ldi YL,_BV(TVTEXT_CURSOR_VISIBLE)
	eor YH,YL
	sts tvtext_flags,YH
	lds YL,tvtext_cursor_flash_period
no_flash_cursor:
	sts tvtext_cursor_flash_timer,YL
	pop r1
	pop r0
	ret

; 4uS of sync.
active_sync:
	cbi _SFR_IO_ADDR(TVTEXT_SYNC_PORT),TVTEXT_SYNC_BIT	
	ldi YL,lo8(cycles_us(t_active_sync))
	ldi YH,hi8(cycles_us(t_active_sync))
	ldi ZL,lo8(pm(active_back_porch))
	ldi ZH,hi8(pm(active_back_porch))
	jmp exit_timer1_compa_vect

; 4uS of back porch.
active_back_porch:
	sbi _SFR_IO_ADDR(TVTEXT_SYNC_PORT),TVTEXT_SYNC_BIT
	; Increment the scanline counter.
	lds YL,tvtext_scanline_counter+0
	inc YL
	sts tvtext_scanline_counter+0,YL
	brne 1f
	lds YH,tvtext_scanline_counter+1
	inc YH
	sts tvtext_scanline_counter+1,YH
1:	ldi YL,lo8(cycles_us(t_active_back_porch))
	ldi YH,hi8(cycles_us(t_active_back_porch))
	ldi ZL,lo8(pm(active_display))
	ldi ZH,hi8(pm(active_display))
	jmp exit_timer1_compa_vect

; 56uS of active display.
active_display:
	sbi _SFR_IO_ADDR(TVTEXT_SYNC_PORT),TVTEXT_SYNC_BIT

	; If TVTEXT_SKIP_ALTERNATE_ROWS is switched on, check if we need to skip.
	.if TVTEXT_SKIP_ALTERNATE_ROWS
	lds ZH,tvtext_scanline_counter+0 ; [2]
	sbrc ZH,0                        ; [2/1]
	jmp skip_output_scanline         ; [2]
	.endif

	; Are we above or below the active display part?
	lds ZH,tvtext_scanline_counter+1
	or ZH,ZH
	breq 1f
	jmp skip_output_scanline
1:

	push r0
	push r1
	push r_font_data_l
	push r_font_data_h
	push r_characters_left
	push r_font_row
	push XL
	push XH

	lds YL,cursor_cell+0 ; [2]
	lds YH,cursor_cell+1 ; [2]
	ld r0,Y              ; [1]
	push r0              ; [1]
	lds r0,cursor_char   ; [2]
	st Y,r0              ; [2]

	; Delay loop to ensure font data appears on-screen:
	.if TVTEXT_SKIP_ALTERNATE_ROWS
	rjmp .     ; [2]
	ldi XL,7   ; [2]
	.else
	ldi XL,9   ; [2]
	.endif
1:	dec XL     ; [1]
	brne 1b    ; [2/1]

	; Get font data offset.
	ldi r_font_data_l,lo8(tvtext_font_data)
	ldi r_font_data_h,hi8(tvtext_font_data)

	; Offset the font data by the scanline number.
	lds XL,tvtext_scanline_counter+0
	lsr XL
	andi XL,7
	clr XH
	add r_font_data_l,XL
	adc r_font_data_h,XH

	; Calculate the position in the text buffer: scanline / 16 * 32 -> (scanline & ~15) * 2
	lds XL,tvtext_scanline_counter+0
	andi XL,~15
	lds XH,tvtext_scanline_counter+1
	lsl XL
	rol XH
	movw r0,XL
	ldi XL,lo8(tvtext_buffer)
	ldi XH,hi8(tvtext_buffer)
	add XL,r0
	adc XH,r1

	ldi ZL, 8             ; [1]
	ld r0, X+             ; [2]
	mul r0, ZL            ; [2] Multiply character index by 8.
	add r0, r_font_data_l ; [1] \_ Offset font data row by scanline number.
	adc r1, r_font_data_h ; [1] /
	movw ZL, r0           ; [1]

	; Clock out a scanline!
	ldi r_characters_left, 192/6                      ; [1]

output_scanline_loop:

	; (1)
	lpm r_font_row, Z                                 ; [3]
	out _SFR_IO_ADDR(TVTEXT_PICTURE_PORT), r_font_row ; [1]
	ldi ZL, 8                                         ; [1]

	; (2)
	ld r0, X+                                         ; [2]
	rol r_font_row                                    ; [1]
	out _SFR_IO_ADDR(TVTEXT_PICTURE_PORT), r_font_row ; [1]
	nop                                               ; [1]

	; (3)
	mul r0, ZL                                        ; [2] Multiply character index by 8.
	rol r_font_row                                    ; [1]
	out _SFR_IO_ADDR(TVTEXT_PICTURE_PORT), r_font_row ; [1]
;	nop                                               ; [1]

	; (4)
	add r0, r_font_data_l                             ; [1] \_ Offset font data row by scanline number.
	adc r1, r_font_data_h                             ; [1] /
	rol r_font_row                                    ; [1]
	out _SFR_IO_ADDR(TVTEXT_PICTURE_PORT), r_font_row ; [1]
	nop                                               ; [1]

	; (5)
	movw ZL, r0                                       ; [1]
	nop                                               ; [1]
	rol r_font_row                                    ; [1]
	out _SFR_IO_ADDR(TVTEXT_PICTURE_PORT), r_font_row ; [1]
	nop                                               ; [1]

	; (6)
	rol r_font_row                                    ; [1]
	dec r_characters_left                             ; [1]
	out _SFR_IO_ADDR(TVTEXT_PICTURE_PORT), r_font_row ; [1]
	brne output_scanline_loop                         ; [2/1]
	nop                                               ; [1]

	clr r_font_row                                    ; [1]
	out _SFR_IO_ADDR(TVTEXT_PICTURE_PORT), r_font_row ; [1]

	; Restore the old character under the cursor.
	pop r0  ; [2]
	st Y,r0 ; [2]

	pop XH
	pop XL
	pop r_font_row
	pop r_characters_left
	pop r_font_data_h
	pop r_font_data_l
	pop r1
	pop r0

skip_output_scanline:

	; Are we done?
	lds YL,scan_region_repeat+0
	lds YH,scan_region_repeat+1
	inc YL
	sts scan_region_repeat+0,YL
	brne active_display_not_finished
	inc YH
	sts scan_region_repeat+1,YH
	brne active_display_not_finished
	
active_display_finished:
	ldi ZL,lo8(pm(short_vsync_sync))
	ldi ZH,hi8(pm(short_vsync_sync))
	ldi YL,6
	sts scan_region_repeat,YL
	jmp active_display_exit
active_display_not_finished:
	ldi ZL,lo8(pm(active_sync))
	ldi ZH,hi8(pm(active_sync))
active_display_exit:
	ldi YL,lo8(cycles_us(t_active_picture))
	ldi YH,hi8(cycles_us(t_active_picture))
	jmp exit_timer1_compa_vect
